package widebase.stream.handler.rq

import akka.actor. { Actor, ActorRef, ActorSystem, Props }

import java.io. { PrintWriter, StringWriter }

import net.liftweb.common.Logger

import org.jboss.netty.channel. {

  Channels,
  ChannelHandlerContext,
  ChannelStateEvent,
  ExceptionEvent,
  MessageEvent,
  SimpleChannelUpstreamHandler

}

import org.jboss.netty.handler.timeout. {

  ReadTimeoutException,
  WriteTimeoutException

}

import widebase.db.table.Table

import widebase.stream.codec. {

  DoneMessage,
  ForbiddenMessage,
  LoginGrantedMessage,
  LoginRequiredMessage,
  UnauthorizedMessage
}

import widebase.stream.codec.rq. {

  EventMessage,
  RollbackMessage,
  TableMessage

}

/** Handles consumer.
 *
 * @param listener of consumer
 *
 * @author myst3r10n
 */
class ConsumerHandler(listener: RecordListener)
  extends SimpleChannelUpstreamHandler
   with Logger {

  /** Holds actors. */
  protected var actors: ActorSystem = _

  /** Read messages. */
  protected var reader: ActorRef = _

  override def channelClosed(ctx: ChannelHandlerContext, evt: ChannelStateEvent) {

    actors.shutdown

  }

  override def channelOpen(ctx: ChannelHandlerContext, evt: ChannelStateEvent) {

    actors = ActorSystem()
    reader = actors.actorOf(Props(new Reader))

  }

  override def exceptionCaught(ctx: ChannelHandlerContext, evt: ExceptionEvent) {

    evt.getCause match {

      case _: ReadTimeoutException =>
        error("Read timeout")
        evt.getChannel.close

      case _: WriteTimeoutException =>
        error("Write timeout")
        evt.getChannel.close

      case _ =>
        val message = new StringWriter
        val printer = new PrintWriter(message)
        evt.getCause.printStackTrace(printer)
        error(message.toString)
        evt.getChannel.close

    }
  }

  override def messageReceived(ctx: ChannelHandlerContext, evt: MessageEvent) {

    evt.getMessage match {

      case message: DoneMessage => Channels.fireMessageReceived(ctx, message)
      case message: ForbiddenMessage => Channels.fireMessageReceived(ctx, message)
      case message: LoginGrantedMessage => Channels.fireMessageReceived(ctx, message)
      case message: LoginRequiredMessage => Channels.fireMessageReceived(ctx, message)
      case message: EventMessage => reader ! message
      case message: RollbackMessage => reader ! message
      case message: TableMessage => reader ! message
      case message: UnauthorizedMessage => Channels.fireMessageReceived(ctx, message)
      case message => error("Unfamiliar with message: " + message)

    }
  }

  protected class Reader extends Actor {

    def receive = {

      case message: EventMessage => listener.receive(message.text)
      case message: RollbackMessage => listener.receive(message.records, message.partition)
      case message: TableMessage => listener.receive(Table.fromBytes(message.bytes))

    }
  }
}

